from __future__ import annotations

from enum import Enum
from typing import Any, List, Optional, Type, Union

import requests
from pydantic import BaseModel, Field

from steamship import MimeTypes, SteamshipError
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import DeleteRequest, IdentifierRequest, Request
from steamship.base.response import Response
from steamship.data.tags.tag import Tag
from steamship.data.tags.tag_constants import ChatTag, DocTag, RoleTag, TagValueKey


class BlockQueryRequest(Request):
    tag_filter_query: str


class BlockUploadType(str, Enum):
    FILE = "file"  # A file uploaded as bytes or a string
    BLOCKS = "blocks"  # Blocks are sent to create a file
    URL = "url"  # content will be fetched from a URL
    NONE = "none"  # No upload; plain text only


def get_tag_value_key(
    tags: Optional[List[Tag]], key: str, kind: Optional[str] = None, name: Optional[str] = None
) -> Optional[any]:
    """Iterates through a list of tags and returns the first that contains the provided Kind/Name/ValueKey."""
    for tag in tags or []:
        if (kind is None or tag.kind == kind) and (name is None or tag.name == name):
            return (tag.value or {}).get(key)
    return None


class Block(CamelModel):
    """A Block is a chunk of content within a File. It can be plain text content, image content,
    video content, etc. If the content is not text, the text value may be the empty string
    for backwards compatibility.
    """

    client: Client = Field(None, exclude=True)
    id: str = None
    file_id: str = None
    text: str = None
    tags: Optional[List[Tag]] = []
    index_in_file: Optional[int] = Field(alias="index")
    mime_type: Optional[MimeTypes]
    public_data: bool = False

    url: Optional[
        str
    ] = None  # Only for creation of blocks; used to fetch content from a public URL.
    content_url: Optional[
        str
    ] = None  # For overriding the URL of the raw data for ephemeral blocks. Setting this will have no effect
    upload_type: Optional[
        BlockUploadType
    ] = None  # for returning Blocks as the result of a generate request
    upload_bytes: Optional[
        bytes
    ] = None  # ONLY for returning Blocks as the result of a generate request. Will not be set when receiving blocks from the server. See raw()

    class ListRequest(Request):
        file_id: str = None

    class ListResponse(Response):
        blocks: List[Block] = []

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["block"] if "block" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def get(
        client: Client,
        _id: str = None,
    ) -> Block:
        return client.post(
            "block/get",
            IdentifierRequest(id=_id),
            expect=Block,
        )

    @staticmethod
    def create(
        client: Client,
        file_id: str,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
        public_data: bool = False,
    ) -> Block:
        """
        Create a new Block within a File specified by file_id.

        You can create a Block in several ways:
        - Providing raw text as the text parameter;
        - Providing the content of the block as string or bytes;
        - Providing a publicly accessible URL where the content is stored.

        """

        if content is not None and url is not None:
            raise SteamshipError("May provide content or URL, but not both when creating a Block")

        if content is not None:
            upload_type = BlockUploadType.FILE
        elif url is not None:
            upload_type = BlockUploadType.URL
        else:
            upload_type = BlockUploadType.NONE

        req = {
            "fileId": file_id,
            "text": text,
            "tags": [t.dict(by_alias=True) for t in tags] if tags else [],
            "url": url,
            "mimeType": mime_type,
            "uploadType": upload_type,
            "publicData": public_data,
        }

        file_data = (
            ("file-part", content, "multipart/form-data")
            if upload_type == BlockUploadType.FILE
            else None
        )

        return client.post(
            "block/create",
            req,
            expect=Block,
            file=file_data,
        )

    def delete(self) -> Block:
        return self.client.post(
            "block/delete",
            DeleteRequest(id=self.id),
            expect=Tag,
        )

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
    ) -> BlockQueryResponse:
        req = BlockQueryRequest(tag_filter_query=tag_filter_query)
        res = client.post(
            "block/query",
            payload=req,
            expect=BlockQueryResponse,
        )
        return res

    def index(self, embedding_plugin_instance: Any = None):
        """Index this block."""
        tags = [
            Tag(
                text=self.text,
                file_id=self.file_id,
                block_id=self.id,
                kind="block",
                start_idx=0,
                end_idx=len(self.text),
            )
        ]
        return embedding_plugin_instance.insert(tags)

    def raw(self):
        if self.content_url is not None:
            return requests.get(self.content_url).content
        else:
            return self.client.post(
                "block/raw",
                payload={
                    "id": self.id,
                },
                raw_response=True,
            )

    def set_public_data(self, public_data: bool):
        """Set the public_data flag on this Block. If this object already exists server-side, update the flag."""
        self.public_data = public_data
        if self.client is not None and self.id is not None:
            req = {
                "id": self.id,
                "publicData": self.public_data,
            }
            return self.client.post("block/update", payload=req, expect=Block)

    def is_text(self) -> bool:
        """Return whether this is a text Block."""
        return self.mime_type == MimeTypes.TXT or (self.mime_type is None and self.text)

    def is_image(self):
        """Return whether this is an image Block."""
        return self.mime_type in [MimeTypes.PNG, MimeTypes.JPG, MimeTypes.GIF, MimeTypes.TIFF]

    def is_audio(self):
        """Return whether this is an audio Block."""
        return self.mime_type in [MimeTypes.MP3, MimeTypes.MP4_AUDIO, MimeTypes.WEBM_AUDIO]

    def is_video(self):
        """Return whether this is a video Block."""
        return self.mime_type in [MimeTypes.MP4_VIDEO, MimeTypes.WEBM_VIDEO]

    @property
    def raw_data_url(self) -> Optional[str]:
        """Return a URL at which the data content of this Block can be accessed.  If public_data is True,
        this content can be accessed without an API key.
        """
        if self.client is not None:
            return f"{self.client.config.api_base}block/{self.id}/raw"
        else:
            return None

    @property
    def chat_role(self) -> Optional[RoleTag]:
        return get_tag_value_key(
            self.tags, TagValueKey.STRING_VALUE, kind=DocTag.CHAT, name=ChatTag.ROLE
        )

    def set_chat_role(self, role: RoleTag):
        return self._one_time_set_tag(
            tag_kind=DocTag.CHAT, tag_name=ChatTag.MESSAGE_ID, string_value=role.value
        )

    @property
    def message_id(self) -> str:
        return get_tag_value_key(
            self.tags, TagValueKey.STRING_VALUE, kind=DocTag.CHAT, name=ChatTag.MESSAGE_ID
        )

    def set_message_id(self, message_id: str):
        return self._one_time_set_tag(
            tag_kind=DocTag.CHAT, tag_name=ChatTag.MESSAGE_ID, string_value=message_id
        )

    @property
    def chat_id(self) -> str:
        return get_tag_value_key(
            self.tags, TagValueKey.STRING_VALUE, kind=DocTag.CHAT, name=ChatTag.CHAT_ID
        )

    def set_chat_id(self, chat_id: str):
        return self._one_time_set_tag(
            tag_kind=DocTag.CHAT, tag_name=ChatTag.CHAT_ID, string_value=chat_id
        )

    def _one_time_set_tag(self, tag_kind: str, tag_name: str, string_value: str):
        existing = get_tag_value_key(
            self.tags, TagValueKey.STRING_VALUE, kind=tag_kind, name=tag_name
        )

        if existing is not None:
            if existing == string_value:
                return  # No action necessary
            else:
                raise SteamshipError(
                    message=f"Block {self.id} already has an existing {tag_kind}/{tag_name} with value {existing}. Unable to set to {string_value}"
                )

        if self.client and self.id:
            tag = Tag.create(
                self.client,
                file_id=self.file_id,
                block_id=self.id,
                kind=tag_kind,
                name=tag_name,
                value={TagValueKey.STRING_VALUE: string_value},
            )
        else:
            tag = Tag(
                kind=tag_kind,
                name=tag_name,
                value={TagValueKey.STRING_VALUE: string_value},
            )

        self.tags.append(tag)

    def as_llm_input(self) -> str:
        if self.is_text():
            return self.text
        else:
            return f"Block({self.id})"


class BlockQueryResponse(Response):
    blocks: List[Block]


Block.ListResponse.update_forward_refs()
Block.update_forward_refs()



import pytest
from steamship_tests import TEST_ASSETS_PATH

from steamship import File, MimeTypes, Steamship
from steamship.experimental import blockify

# @pytest.mark.usefixtures("client")
# def test_blockify_youtube(client: Steamship):
#     file = scrape(client, "https://www.youtube.com/watch?v=LXDZ6aBjv_I")
#     assert file.id is not None
#     content = file.raw()
#     assert content is not None
#     task = blockify(file)
#     task.wait()
#
#     file2 = file.refresh()
#     assert len(file2.blocks) > 0
#
#     file.delete()


@pytest.mark.usefixtures("client")
def test_blockify_pdf(client: Steamship):
    palm_tree_path = TEST_ASSETS_PATH / "test.pdf"

    with palm_tree_path.open("rb") as f:
        palm_bytes = f.read()

    file = File.create(client, content=palm_bytes, mime_type=MimeTypes.PDF)
    assert file.mime_type == MimeTypes.PDF
    assert file.id is not None
    content = file.raw()
    assert content is not None
    task = blockify(file)
    task.wait()

    file2 = file.refresh()
    assert len(file2.blocks) > 0

    blocks = file.blocks
    assert blocks
    assert len(blocks) == 2
    assert blocks[0].text == "This is the ﬁrst page\n"
    assert blocks[1].text == "This is the second page"

    file.delete()



import logging
import re
from typing import Dict, List, Optional

from steamship import Block, Steamship
from steamship.agents.schema import Action, AgentContext, FinishAction, OutputParser, Tool


class ReACTOutputParser(OutputParser):
    """Parse LLM output expecting structure matching ReACTAgent default prompt."""

    tools_lookup_dict: Optional[Dict[str, Tool]] = None

    def __init__(self, **kwargs):
        tools_lookup_dict = {tool.name: tool for tool in kwargs.pop("tools", None)}
        super().__init__(tools_lookup_dict=tools_lookup_dict, **kwargs)

    def parse(self, text: str, context: AgentContext) -> Action:
        if text.endswith("No"):
            raise RuntimeError(f"Could not parse LLM output: `{text}`")

        if "AI:" in text:
            return FinishAction(
                output=ReACTOutputParser._blocks_from_text(context.client, text), context=context
            )

        regex = r"Action: (.*?)[\n]*Action Input: (.*)"
        match = re.search(regex, text)
        if not match:
            logging.warning(f"Bad agent response ({text}). Returning results directly to the user.")
            # TODO: should this be the case?  If we are off-base should we just return what we have?
            return FinishAction(
                output=ReACTOutputParser._blocks_from_text(context.client, text), context=context
            )
        action = match.group(1)
        action_input = match.group(2).strip()
        tool = self.tools_lookup_dict.get(action.strip(), None)
        if tool is None:
            raise RuntimeError(
                f"Could not find tool from action: `{action}`. Known tools: {self.tools_lookup_dict.keys()}"
            )
        return Action(
            tool=tool,
            input=[Block(text=action_input)],
            context=context,
        )

    @staticmethod
    def _blocks_from_text(client: Steamship, text: str) -> List[Block]:
        last_response = text.split("AI:")[-1].strip()

        block_id_regex = r"(?:(?:\[|\()?Block)?\(?([A-F0-9]{8}\-[A-F0-9]{4}\-[A-F0-9]{4}\-[A-F0-9]{4}\-[A-F0-9]{12})\)?(?:(\]|\)))?"
        remaining_text = last_response
        result_blocks: List[Block] = []
        while remaining_text is not None and len(remaining_text) > 0:
            match = re.search(block_id_regex, remaining_text)
            if match:
                pre_block_text = ReACTOutputParser._remove_block_prefix(
                    candidate=remaining_text[0 : match.start()]
                )
                if len(pre_block_text) > 0:
                    result_blocks.append(Block(text=pre_block_text))
                result_blocks.append(Block.get(client, _id=match.group(1)))
                remaining_text = ReACTOutputParser._remove_block_suffix(
                    remaining_text[match.end() :]
                )
            else:
                result_blocks.append(Block(text=remaining_text))
                remaining_text = ""

        return result_blocks

    @staticmethod
    def _remove_block_prefix(candidate: str) -> str:
        removed = candidate
        if removed.endswith("(Block") or removed.endswith("[Block"):
            removed = removed[len("Block") + 1 :]
        elif removed.endswith("Block"):
            removed = removed[len("Block") :]
        return removed

    @staticmethod
    def _remove_block_suffix(candidate: str) -> str:
        removed = candidate
        if removed.startswith(")") or removed.endswith("]"):
            removed = removed[1:]
        return removed



from typing import Any, Dict

from steamship import Steamship
from steamship.invocable import Invocable


class BadPackage(Invocable):
    """Bad Package should not start when deployed."""

    def __init__(self, client: Steamship = None, config: Dict[str, Any] = None):
        super().__init__(client, config)
        raise RuntimeError("this should fail immediately")



from enum import Enum


class Skill(str, Enum):
    ENTITIES = "entities"
    SUMMARY = "summary"
    SENTIMENTS = "sentiments"
    EMOTIONS = "emotions"
    TOPICS = "topics"
    HIGHLIGHTS = "highlights"
    KEYWORDS = "keywords"



import logging
import tempfile
from typing import Any, Dict, Optional

import requests

from steamship import Block, Steamship, SteamshipError
from steamship.agents.schema import Metadata
from steamship.experimental.transports.transport import Transport

API_BASE = "https://api.telegram.org/bot"


class TelegramTransport(Transport):
    """Experimental base class to encapsulate a Telegram communication channel."""

    api_root: str
    bot_token: str

    def __init__(self, bot_token: str, client: Steamship):
        super().__init__(client=client)
        self.api_root = f"{API_BASE}{bot_token}"
        self.bot_token = bot_token

    def _instance_init(self, *args, **kwargs):
        if "webhook_url" not in kwargs:
            raise SteamshipError(
                message="Please provide the `webhook_url` method in the TelegramTransport instance_init kwargs"
            )

        webhook_url = kwargs["webhook_url"]
        logging.info(f"Setting Telegram webhook URL: {webhook_url}")
        logging.info(f"Post is to {self.api_root}/setWebhook")

        response = requests.get(
            f"{self.api_root}/setWebhook",
            params={
                "url": webhook_url,
                "allowed_updates": ["message"],
                "drop_pending_updates": True,
            },
        )

        if not response.ok:
            raise SteamshipError(
                f"Could not set webhook for bot. Webhook URL was {webhook_url}. Telegram response message: {response.text}"
            )

    def _instance_deinit(self, *args, **kwargs):
        """Unsubscribe from Telegram updates."""
        requests.get(f"{self.api_root}/deleteWebhook")

    def _send(self, blocks: [Block], metadata: Metadata):
        """Send a response to the Telegram chat."""
        for block in blocks:
            chat_id = block.chat_id
            if block.is_text() or block.text:
                params = {"chat_id": int(chat_id), "text": block.text}
                requests.get(f"{self.api_root}/sendMessage", params=params)
            elif block.is_image() or block.is_audio() or block.is_video():
                if block.is_image():
                    suffix = "sendPhoto"
                    key = "photo"
                elif block.is_audio():
                    suffix = "sendAudio"
                    key = "audio"
                elif block.is_video():
                    suffix = "sendVideo"
                    key = "video"

                _bytes = block.raw()
                with tempfile.TemporaryFile(mode="r+b") as temp_file:
                    temp_file.write(_bytes)
                    temp_file.seek(0)
                    resp = requests.post(
                        url=f"{self.api_root}/{suffix}?chat_id={chat_id}",
                        files={key: temp_file},
                    )
                    if resp.status_code != 200:
                        logging.error(f"Error sending message: {resp.text} [{resp.status_code}]")
                        raise SteamshipError(
                            f"Message not sent to chat {chat_id} successfully: {resp.text}"
                        )
            else:
                logging.error(
                    f"Telegram transport unable to send a block of MimeType {block.mime_type}"
                )

    def _info(self) -> dict:
        """Fetches info about this bot."""
        resp = requests.get(f"{self.api_root}/getMe").json()
        logging.info(f"/info: {resp}")
        return {"telegram": resp.get("result")}

    def _get_file(self, file_id: str) -> Dict[str, Any]:
        return requests.get(f"{self.api_root}/getFile", params={"file_id": file_id}).json()[
            "result"
        ]

    def _get_file_url(self, file_id: str) -> str:
        return f"https://api.telegram.org/file/bot{self.bot_token}/{self._get_file(file_id)['file_path']}"

    def _download_file(self, file_id: str):
        result = requests.get(self._get_file_url(file_id))
        if result.status_code != 200:
            raise Exception("Download file", result)

        return result.content

    def _parse_inbound(self, payload: dict, context: Optional[dict] = None) -> Optional[Block]:
        """Parses an inbound Telegram message."""

        chat = payload.get("chat")
        if chat is None:
            raise SteamshipError(f"No `chat` found in Telegram message: {payload}")

        chat_id = chat.get("id")
        if chat_id is None:
            raise SteamshipError(f"No 'chat_id' found in Telegram message: {payload}")

        if not isinstance(chat_id, int):
            raise SteamshipError(
                f"Bad 'chat_id' found in Telegram message: ({chat_id}). Should have been an int."
            )

        message_id = payload.get("message_id")
        if message_id is None:
            raise SteamshipError(f"No 'message_id' found in Telegram message: {payload}")

        if not isinstance(message_id, int):
            raise SteamshipError(
                f"Bad 'message_id' found in Telegram message: ({message_id}). Should have been an int"
            )

        if "voice" in payload:
            file_id = payload.get("voice").get("file_id")
            voice_file_url = self._get_file_url(file_id)
            return Block(
                text=payload.get("text"),
                url=voice_file_url,
                chat_id=str(chat_id),
                message_id=str(message_id),
            )

        # Some incoming messages (like the group join message) don't have message text.
        # Rather than throw an error, we just don't return a Block.
        message_text = payload.get("text")
        if message_text is not None:
            result = Block(text=message_text)
            result.set_chat_id(str(chat_id))
            result.set_message_id(str(message_id))
            return result
        else:
            return None



"""Tool for generating images."""
from steamship.agents.tools import ImageGeneratorTool
from steamship.utils.repl import ToolREPL


class DalleTool(ImageGeneratorTool):
    """Tool to generate images from text using OpenAI's DALL-E."""

    name: str = "DalleTool"
    human_description: str = "Generates an image from text."
    agent_description = (
        "Used to generate images from text prompts. Only use if the user has asked directly for an "
        "image. When using this tool, the input should be a plain text string that describes, "
        "in detail, the desired image."
    )
    generator_plugin_handle: str = "dall-e"
    generator_plugin_config: dict = {"n": 1, "size": "256x256"}


if __name__ == "__main__":
    ToolREPL(DalleTool()).run()



import pytest

from steamship import Block, File, Steamship
from steamship.agents.react import ReACTOutputParser


@pytest.mark.usefixtures("client")
def test_parse_output(client: Steamship):

    file = File.create(client, blocks=[Block(text="test"), Block(text="Another test")])
    block_id = file.blocks[0].id
    block_2_id = file.blocks[1].id

    example1 = f"some text [Block({block_id})] some more text"

    parsed_blocks1 = ReACTOutputParser._blocks_from_text(client, example1)
    assert len(parsed_blocks1) == 3
    assert parsed_blocks1[0].text == "some text "
    assert parsed_blocks1[1].id == block_id
    assert parsed_blocks1[2].text == " some more text"

    example2 = f"some text [Block({block_id})] some more text {block_2_id} even more text"
    parsed_blocks2 = ReACTOutputParser._blocks_from_text(client, example2)
    assert len(parsed_blocks2) == 5
    assert parsed_blocks2[0].text == "some text "
    assert parsed_blocks2[1].id == block_id
    assert parsed_blocks2[2].text == " some more text "
    assert parsed_blocks2[3].id == block_2_id
    assert parsed_blocks2[4].text == " even more text"

    example3 = f"some text [Block({block_id})] some more text {block_2_id}"
    parsed_blocks3 = ReACTOutputParser._blocks_from_text(client, example3)
    assert len(parsed_blocks3) == 4
    assert parsed_blocks3[0].text == "some text "
    assert parsed_blocks3[1].id == block_id
    assert parsed_blocks3[2].text == " some more text "
    assert parsed_blocks3[3].id == block_2_id



from .dalle import DalleTool
from .google_image_search import GoogleImageSearchTool
from .stable_diffusion import StableDiffusionTool

__all__ = [
    "DalleTool",
    "GoogleImageSearchTool",
    "StableDiffusionTool",
]



from typing import Any, Callable, Dict, List

from steamship import Block, Steamship, Tag
from steamship.agents.memory import ChatHistory
from steamship.agents.schema.action import Action

Metadata = Dict[str, Any]
EmitFunc = Callable[[List[Block], Metadata], None]


class AgentContext:
    """AgentContext contains all relevant information about a particular execution of an Agent. It is used by the
    AgentService to manage execution history as well as store/retrieve information and metadata that will be used
    in the process of an agent execution.
    """

    # A steamship-assigned ID for this memory object
    @property
    def id(self) -> str:
        return self.chat_history.file.id

    metadata: Metadata = {}
    """Allows storage of arbitrary information that may be useful for agents and tools."""

    client: Steamship
    """Provides workspace-specific utilities for the agents and tools."""

    chat_history: ChatHistory
    """Record of user-package messages. It records user submitted queries/prompts and the final
    agent-driven answer sent in response to those queries/prompts. It does NOT record any chat
    history related to agent execution and action selection."""

    completed_steps: List[Action] = []
    """Record of agent-selected Actions and their outputs. This provides an ordered look at the
    execution sequence for this context."""

    # todo: in the future, this could be a set of callbacks like onError, onComplete, ...
    emit_funcs: List[EmitFunc] = []
    """Called when an agent execution has completed. These provide a way for the AgentService
    to return the result of an agent execution to the package that requested the agent execution."""

    @staticmethod
    def get_or_create(
        client: Steamship,
        context_keys: Dict[str, str],
        tags: List[Tag] = None,
    ):
        history = ChatHistory.get_or_create(client, context_keys, tags)
        context = AgentContext()
        context.chat_history = history
        context.client = client
        context.completed_steps = []
        return context



import logging
from typing import Any, Dict, Optional

from steamship.base.client import Client
from steamship.base.error import SteamshipError
from steamship.data.plugin.plugin_instance import CreatePluginInstanceRequest, PluginInstance
from steamship.data.tags.tag_constants import TagKind, TagValueKey


class PromptGenerationPluginInstance(PluginInstance):
    """An instance of a configured prompt completion service such as GPT-3.

    The `generate` method synchronously invokes the prompt against a set of variables that parameterize it.
    The return value is a single string.

    Example Usage:
       llm = Steamship.use('prompt-generation-default', config={ "temperature": 0.9 })
       PROMPT = "Greet {name} as if he were a {relation}."
       greeting = llm.generate(PROMPT, {"name": "Ted", "relation": "old friend"})
    """

    def generate(
        self, prompt: str, variables: Optional[Dict] = None, clean_output: bool = True
    ) -> str:
        """Complete the provided prompt, interpolating any variables."""

        # Interpolate the prompt with Python formatting semantics. If no variables provided, supply an empty dict.
        try:
            prompt_text = prompt.format(**(variables or {}))
        except KeyError as e:
            raise SteamshipError(
                message="Some variables in the prompt template were not provided.", error=e
            )

        # This requests generation from the parameterized prompt. Tagging with our prompt generator
        # plugin will result in a new tag that contains the generated output.
        tag_task = self.tag(doc=prompt_text)

        # We `wait()` because generation of text is done asynchronously and may take a few moments
        # (somewhat depending on the complexity of your prompt).
        tag_task.wait()

        # Here, we iterate through the content blocks associated with a file
        # as well as any tags on that content to find the generated text.
        #
        # The Steamship data model provides flexible content organization,
        # storage, and lookup. Read more about the data model via:
        # https://docs.steamship.com/workspaces/data_model/index.html
        try:
            for text_block in tag_task.output.file.blocks:
                for block_tag in text_block.tags:
                    if block_tag.kind == TagKind.GENERATION:
                        generation = block_tag.value[TagValueKey.STRING_VALUE]
                        if clean_output:
                            return self._clean_output(generation)
                        else:
                            return generation
        except Exception as e:
            logging.error(
                "generate() got unexpected response shape back. This suggests an error rather an merely an empty response."
            )
            logging.exception(e)
            raise e
        return ""

    @staticmethod
    def create(
        client: Client,
        plugin_id: str = None,
        plugin_handle: str = None,
        plugin_version_id: str = None,
        plugin_version_handle: str = None,
        handle: str = None,
        fetch_if_exists: bool = True,
        config: Dict[str, Any] = None,
    ) -> "PromptGenerationPluginInstance":
        """Create a plugin instance

        When handle is empty the engine will automatically assign one
        fetch_if_exists controls whether we want to re-use an existing plugin instance or not."""
        req = CreatePluginInstanceRequest(
            handle=handle,
            plugin_id=plugin_id,
            plugin_handle=plugin_handle,
            plugin_version_id=plugin_version_id,
            plugin_version_handle=plugin_version_handle,
            fetch_if_exists=fetch_if_exists,
            config=config,
        )

        return client.post(
            "plugin/instance/create", payload=req, expect=PromptGenerationPluginInstance
        )

    def _clean_output(self, text: str):
        """Remove any leading/trailing whitespace and partial sentences.

        This assumes that your generated output will include consistent punctuation. You may
        want to alter this method to better fit the format of your generated text.
        """
        last_punc = -1
        for i, c in enumerate(reversed(text)):
            if c in '.!?"':
                last_punc = len(text) - i
                break
        if last_punc != -1:
            result = text[: last_punc + 1]
        else:
            result = text
        return result.strip()



import contextlib

from steamship_tests import TEST_ASSETS_PATH

from steamship import File, Steamship


@contextlib.contextmanager
def upload_file(client: Steamship, test_asset_filename: str):
    full_path = TEST_ASSETS_PATH / test_asset_filename
    file = File.create(client, content=full_path.open("rb").read())
    assert file is not None
    assert file.id is not None
    yield file
    file.delete()



import json
import logging
from pathlib import Path
from typing import Optional

import requests

from steamship import SteamshipError
from steamship.utils.url import apply_localstack_url_fix


def url_to_json(url: str) -> any:
    """
    Downloads the Signed URL and returns the contents as JSON.
    """
    bytes = url_to_bytes(url)
    json_string = bytes.decode("utf8")
    return json.loads(json_string)


def url_to_bytes(url: str) -> bytes:
    """
    Downloads the Signed URL and returns the contents as bytes.

    This is a helper function to consolidate Steamship Client URL fetching to ensure a single point of handling for:
      * Error messages
      * Any required manipulations for URL signed URLs
      * Any required manipulations for localstack-based environments

    Note that the base API Client does not use this method on purpose: in the event of error code, it inspects the
    contents of the response for a SteamshipError.
    """
    url = apply_localstack_url_fix(url)
    logging.info(f"Downloading: {url}.")

    resp = requests.get(url)
    if resp.status_code != 200:
        # TODO: At least Localstack send to reply with HTTP 200 even if the file isn't found!
        # The full response contains:
        # <Error>
        #     <Code>NoSuchKey</Code>
        #
        # So we **could** check the response text even in the event of 200 but that seems wrong..
        if "<Code>NoSuchKey</Code>" in resp.text:
            raise SteamshipError(
                message=f"The file at signed URL {url} did not exist. HTTP {resp.status_code}. Content: {resp.text}"
            )
        else:
            raise SteamshipError(
                message=f"There was an error downloading from the signed url: {url}. HTTP {resp.status_code}. Content: {resp.text}"
            )
    return resp.content


def download_from_signed_url(url: str, to_file: Path = None) -> Path:
    """
    Downloads the Signed URL to the filename `desired_filename` in a temporary directory on disk.
    """
    content = url_to_bytes(url)

    if not to_file.parent.exists():
        to_file.parent.mkdir(parents=True, exist_ok=True)

    with open(to_file, "wb") as f:
        logging.debug(f"Got contents of: {url}")
        f.write(content)
        logging.debug(f"Wrote contents of: {url} to {to_file}")
    return Path(to_file)


def upload_to_signed_url(url: str, _bytes: Optional[bytes] = None, filepath: Optional[Path] = None):
    """
    Uploads either the bytes or filepath contents to the provided Signed URL.
    """

    url = apply_localstack_url_fix(url)
    if _bytes is not None:
        logging.info(f"Uploading provided bytes to: {url}")
    elif filepath is not None:
        logging.info(f"Uploading file at {filepath} to: {url}")
        with open(filepath, "rb") as f:
            _bytes = f.read()
    else:
        raise SteamshipError(
            message="Unable to upload data to signed URL -- neither a filepath nor bytes were provided.",
            suggestion="Please provide either the `bytes` or the `filepath` argument",
        )

    http_response = requests.put(
        url, data=_bytes, headers={"Content-Type": "application/octet-stream"}
    )

    # S3 returns 204 upon success; we include 200 here for safety.
    if http_response.status_code not in [200, 204]:
        logging.error(f"File upload error. file={filepath}. url= {url}")
        logging.error(f"Status Code: {http_response.status_code}")
        logging.error(f"Response Text: {http_response.text}")
        raise SteamshipError(
            message=f"Unable to upload data to signed URL. Status code: {http_response.status_code}. Status text: {http_response.text}"
        )



"""Common utilities for the Steamship Python API"""
import os
import sys
import traceback
from importlib import machinery
from pathlib import Path

import click

from steamship.base.error import SteamshipError


def find_api_py() -> Path:
    path = Path("src/api.py")
    if not path.exists():
        path = Path("api.py")
        if not path.exists():
            raise SteamshipError("Could not find api.py either in root directory or in src.")
    return path


def get_api_module(path: Path):
    try:
        sys.path.append(str(path.parent.absolute()))

        # load the API module to allow config inspection / generation
        return machinery.SourceFileLoader("api", str(path)).load_module()
    except Exception:
        click.secho(
            "An error occurred while loading your api.py to check configuration parameters. Full stack trace below.",
            fg="red",
        )
        traceback.print_exc()
        click.get_current_context().abort()
        return None


def is_in_replit() -> bool:
    """Returns True if the user appears to be inside Replit."""
    if os.environ.get("REPLIT_CLI") or os.environ.get("REPLIT_DB_URL"):
        return True

    return False



import logging

from steamship.invocable import Invocable, InvocableResponse, post


class PackageWithInstanceInit(Invocable):
    def instance_init(self):
        logging.info(f"Init called with URL: {self.context.invocable_url}")

    @post("was_init_called")
    def was_init_called(self) -> InvocableResponse:
        return InvocableResponse(data="OK")



from enum import Enum
from typing import Optional

from steamship.base.model import CamelModel


class Request(CamelModel):
    pass


class GetRequest(Request):
    id: str = None
    handle: str = None


class CreateRequest(Request):
    id: str = None
    handle: str = None


class UpdateRequest(Request):
    id: str = None
    handle: str = None


class IdentifierRequest(Request):
    id: str = None
    handle: str = None


class SortOrder(str, Enum):
    ASC = "ASC"
    DESC = "DESC"


class ListRequest(Request):
    page_size: Optional[int]
    page_token: Optional[str]
    sort_order: Optional[SortOrder] = SortOrder.DESC


class DeleteRequest(Request):
    id: str



from __future__ import annotations

from typing import List, Optional

from steamship import Block
from steamship.base import SteamshipError
from steamship.base.model import CamelModel
from steamship.utils.signed_urls import url_to_json


class RawBlockAndTagPluginInput(CamelModel):
    blocks: List[Block]
    options: Optional[dict]

    def __init__(self, **kwargs):
        if url := kwargs.get("url"):
            # If `url` was provided, we assume that some or all of the new object's parameterization exists
            # at that location, encoded as JSON. We fetch it, parse as JSON, and fold into the kwarg dict.
            result = url_to_json(url)
            if not isinstance(result, dict):
                raise SteamshipError(
                    message=f"BlockAndTagPluginInput received a URL that resolved to {type(result)}. Needed a `dict`"
                )
            kwargs.update(result)

        super().__init__(**kwargs)



from enum import Enum
from typing import List, Optional

from steamship.base.model import CamelModel


class OperationType(str, Enum):
    CREATE = "create"
    DELETE = "delete"
    RUN = "run"
    TRAIN = "train"
    EXISTS = "exists"
    EXISTS_TRAINED = "existsTrained"


class OperationUnit(str, Enum):
    CHARACTERS = "characters"
    UNITS = "units"
    BYTES = "bytes"
    MS = "ms"  # milliseconds
    TOKENS = "tokens"
    PROMPT_TOKENS = "promptTokens"
    SAMPLED_TOKENS = "sampledTokens"


class UsageReport(CamelModel):
    """This is the report object that a plugin or package can send back to notify the engine how much of something was consumed"""

    # What type of operation was performed
    operation_type: OperationType

    # What the unit of measurement is.
    operation_unit: OperationUnit

    # What the magnitude of measurement is
    operation_amount: int

    # A URL, if known, to an external audit object for this usage event
    audit_url: Optional[str]

    # An ID, if known, to an external record for this usage event
    audit_id: Optional[str]

    @staticmethod
    def run_tokens(tokens: int, audit_url: Optional[str] = None, audit_id: Optional[str] = None):
        return UsageReport(
            operation_type=OperationType.RUN,
            operation_unit=OperationUnit.TOKENS,
            operation_amount=tokens,
            audit_url=audit_url,
            audit_id=audit_id,
        )

    @staticmethod
    def run_units(units: int, audit_url: Optional[str] = None, audit_id: Optional[str] = None):
        return UsageReport(
            operation_type=OperationType.RUN,
            operation_unit=OperationUnit.UNITS,
            operation_amount=units,
            audit_url=audit_url,
            audit_id=audit_id,
        )

    @staticmethod
    def run_characters(
        characters: int, audit_url: Optional[str] = None, audit_id: Optional[str] = None
    ):
        return UsageReport(
            operation_type=OperationType.RUN,
            operation_unit=OperationUnit.CHARACTERS,
            operation_amount=characters,
            audit_url=audit_url,
            audit_id=audit_id,
        )


class PluginOutput(CamelModel):
    """Base class for all types of plugin output, allowing usage reporting"""

    usage: Optional[List[UsageReport]]



"""Tool for generating images."""
from steamship.agents.tools import AudioBlockifierTool
from steamship.utils.repl import ToolREPL


class WhisperSpeechToTextTool(AudioBlockifierTool):
    """Tool to generate audio from text."""

    name: str = "WhisperSpeechToTextTool"
    human_description: str = "Generates text from spoken audio."
    agent_description: str = (
        "Used to generate text from spoken audio at a URL. Only use if the user has asked directly for a an text version of an audio file. "
        "The input is a URL. "
        "The output is the text from that URL."
    )
    blockifier_plugin_handle: str = "whisper-s2t-blockifier"


if __name__ == "__main__":
    print(
        "You can try: https://d3ctxlq1ktw2nl.cloudfront.net/staging/2023-4-14/2ba758be-d152-f80d-4ef0-9c63789e556b.mp3"
    )
    ToolREPL(WhisperSpeechToTextTool()).run()
    # Try on: https://anchor.fm/s/e1369b4c/podcast/play/70381739/https%3A%2F%2Fd3ctxlq1ktw2nl.cloudfront.net%2Fstaging%2F2023-4-14%2Fda96d64d-3c27-3e84-eb63-d5bc55eaf52e.mp3



from __future__ import annotations

from steamship.data.file import File
from steamship.plugin.outputs.plugin_output import PluginOutput


class BlockAndTagPluginOutput(PluginOutput):
    file: File = None



from typing import List, Optional

from steamship import Block
from steamship.agents.service.agent_service import AgentService
from steamship.experimental.package_starters.telegram_bot import TelegramBot


class MultiInheritanceBot(AgentService, TelegramBot):
    def create_response(self, incoming_message: Block) -> Optional[List[Block]]:
        return [Block(text="Yo")]



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Optional

import inflection
from pydantic import HttpUrl, SecretStr

from steamship.base.model import CamelModel
from steamship.cli.login import login
from steamship.utils.utils import format_uri

DEFAULT_WEB_BASE = "https://steamship.com/"
DEFAULT_APP_BASE = "https://steamship.run/"
DEFAULT_API_BASE = "https://api.steamship.com/api/v1/"

ENVIRONMENT_VARIABLES_TO_PROPERTY = {
    "STEAMSHIP_API_KEY": "api_key",
    "STEAMSHIP_API_BASE": "api_base",
    "STEAMSHIP_APP_BASE": "app_base",
    "STEAMSHIP_WEB_BASE": "web_base",
    "STEAMSHIP_WORKSPACE_ID": "workspace_id",
    "STEAMSHIP_WORKSPACE_HANDLE": "workspace_handle",
}
DEFAULT_CONFIG_FILE = Path.home() / ".steamship.json"

# This stops us from including the `client` object in the dict() output, which is fine in a dict()
# but explodes if that dict() is turned into JSON. Sadly the `exclude` option in Pydantic doesn't
# cascade down nested objects, so we have to use this structure to catch all the possible combinations
EXCLUDE_FROM_DICT = {
    "client": True,
    "blocks": {"__all__": {"client": True, "tags": {"__all__": {"client": True}}}},
    "tags": {"__all__": {"client": True}},
}


class Configuration(CamelModel):
    api_key: SecretStr
    api_base: HttpUrl = DEFAULT_API_BASE
    app_base: HttpUrl = DEFAULT_APP_BASE
    web_base: HttpUrl = DEFAULT_WEB_BASE
    workspace_id: str = None
    workspace_handle: str = None
    profile: Optional[str] = None

    def __init__(
        self,
        config_file: Optional[Path] = None,
        **kwargs,
    ):
        # First set the profile
        kwargs["profile"] = profile = kwargs.get("profile") or os.getenv("STEAMSHIP_PROFILE")

        # Then load configuration from a file if provided
        config_dict = self._load_from_file(
            config_file or DEFAULT_CONFIG_FILE,
            profile,
            raise_on_exception=config_file is not None,
        )
        config_dict.update(self._get_config_dict_from_environment())
        kwargs.update({k: v for k, v in config_dict.items() if kwargs.get(k) is None})

        kwargs["api_base"] = format_uri(kwargs.get("api_base"))
        kwargs["app_base"] = format_uri(kwargs.get("app_base"))
        kwargs["web_base"] = format_uri(kwargs.get("web_base"))

        if not kwargs.get("api_key") and not kwargs.get("apiKey"):
            api_key = login(
                kwargs.get("api_base") or DEFAULT_API_BASE,
                kwargs.get("web_base") or DEFAULT_WEB_BASE,
            )
            Configuration._save_api_key_to_file(
                api_key, profile, config_file or DEFAULT_CONFIG_FILE
            )
            kwargs["api_key"] = api_key

        super().__init__(**kwargs)

    @staticmethod
    def _load_from_file(
        file: Path, profile: str = None, raise_on_exception: bool = False
    ) -> Optional[dict]:
        try:
            with file.open() as f:
                config_file = json.load(f)
                if profile:
                    if "profiles" not in config_file or profile not in config_file["profiles"]:
                        raise RuntimeError(f"Profile {profile} requested but not found in {file}")
                    config = config_file["profiles"][profile]
                else:
                    config = config_file
                return {inflection.underscore(k): v for k, v in config.items()}
        except FileNotFoundError:
            if raise_on_exception:
                raise Exception(f"Tried to load configuration file at {file} but it did not exist.")
        except Exception as err:
            if raise_on_exception:
                raise err
        return {}

    @staticmethod
    def _get_config_dict_from_environment():
        """Overrides configuration with environment variables."""
        return {
            property_name: os.getenv(environment_variable_name, None)
            for environment_variable_name, property_name in ENVIRONMENT_VARIABLES_TO_PROPERTY.items()
            if environment_variable_name in os.environ
        }

    @staticmethod
    def _save_api_key_to_file(new_api_key: Optional[str], profile: Optional[str], file_path: Path):
        # Minimally rewrite config file, adding api key
        try:
            with file_path.open() as f:
                config_file = json.load(f)
            if profile:
                if "profiles" not in config_file or profile not in config_file["profiles"]:
                    raise RuntimeError(f"Could not update API key for {profile} in {file_path}")
                config = config_file["profiles"][profile]
            else:
                config = config_file
        except FileNotFoundError:
            config_file = {}
            config = config_file

        config["apiKey"] = new_api_key

        with file_path.open("w") as f:
            json.dump(config_file, f, indent="\t")

    @staticmethod
    def default_config_file_has_api_key() -> bool:
        return Configuration._load_from_file(DEFAULT_CONFIG_FILE).get("api_key") is not None

    @staticmethod
    def remove_api_key_from_default_config():
        Configuration._save_api_key_to_file(None, None, DEFAULT_CONFIG_FILE)



from __future__ import annotations

import time
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import BaseModel, Field

from steamship import SteamshipError
from steamship.base import Task
from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import DeleteRequest, IdentifierRequest, Request
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.invocable_init_status import InvocableInitStatus
from steamship.data.operations.generator import GenerateRequest, GenerateResponse
from steamship.data.operations.tagger import TagRequest, TagResponse
from steamship.data.plugin import (
    HostingCpu,
    HostingEnvironment,
    HostingMemory,
    HostingTimeout,
    HostingType,
)
from steamship.plugin.inputs.export_plugin_input import ExportPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import TrainingParameterPluginInput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import TrainingParameterPluginOutput


class CreatePluginInstanceRequest(Request):
    id: str = None
    plugin_id: str = None
    plugin_handle: str = None
    plugin_version_id: str = None
    plugin_version_handle: str = None
    handle: str = None
    fetch_if_exists: bool = None
    config: Dict[str, Any] = None


SIGNED_URL_EXPORTER_INSTANCE_HANDLE = "signed-url-exporter-1.0"


class PluginInstance(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    handle: str = None
    plugin_id: str = None
    plugin_version_id: str = None
    plugin_handle: Optional[str] = None
    plugin_version_handle: Optional[str] = None
    workspace_id: Optional[str] = None
    user_id: str = None
    config: Dict[str, Any] = None
    hosting_type: Optional[HostingType] = None
    hosting_cpu: Optional[HostingCpu] = None
    hosting_memory: Optional[HostingMemory] = None
    hosting_timeout: Optional[HostingTimeout] = None
    hosting_environment: Optional[HostingEnvironment] = None
    init_status: Optional[InvocableInitStatus] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["pluginInstance"] if "pluginInstance" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        plugin_id: str = None,
        plugin_handle: str = None,
        plugin_version_id: str = None,
        plugin_version_handle: str = None,
        handle: str = None,
        fetch_if_exists: bool = True,
        config: Dict[str, Any] = None,
    ) -> PluginInstance:
        """Create a plugin instance

        When handle is empty the engine will automatically assign one
        fetch_if_exists controls whether we want to re-use an existing plugin instance or not."""
        req = CreatePluginInstanceRequest(
            handle=handle,
            plugin_id=plugin_id,
            plugin_handle=plugin_handle,
            plugin_version_id=plugin_version_id,
            plugin_version_handle=plugin_version_handle,
            fetch_if_exists=fetch_if_exists,
            config=config,
        )

        return client.post("plugin/instance/create", payload=req, expect=PluginInstance)

    @staticmethod
    def get(client: Client, handle: str) -> PluginInstance:
        return client.post(
            "plugin/instance/get", IdentifierRequest(handle=handle), expect=PluginInstance
        )

    def tag(
        self,
        doc: Union[str, File],
    ) -> Task[
        TagResponse
    ]:  # TODO (enias): Should we remove this helper function in favor of always working with files?
        req = TagRequest(
            type="inline",
            file=File(blocks=[Block(text=doc)]) if isinstance(doc, str) else doc,
            plugin_instance=self.handle,
        )
        return self.client.post(
            "plugin/instance/tag",
            req,
            expect=TagResponse,
        )

    def generate(
        self,
        input_file_id: str = None,
        input_file_start_block_index: int = None,
        input_file_end_block_index: Optional[int] = None,
        input_file_block_index_list: Optional[List[int]] = None,
        text: Optional[str] = None,
        # bytes: Optional[bytes] = None, [Not yet implemented]
        block_query: Optional[str] = None,
        # url: Optional[str] = None, [Not yet implemented]
        append_output_to_file: bool = False,
        output_file_id: Optional[str] = None,
        make_output_public: Optional[bool] = None,
        options: Optional[dict] = None,
    ) -> Task[GenerateResponse]:
        """See GenerateRequest for description of parameter options"""
        req = GenerateRequest(
            plugin_instance=self.handle,
            input_file_id=input_file_id,
            input_file_start_block_index=input_file_start_block_index,
            input_file_end_block_index=input_file_end_block_index,
            input_file_block_index_list=input_file_block_index_list,
            text=text,
            # bytes=bytes,
            block_query=block_query,
            # url=url,
            append_output_to_file=append_output_to_file,
            output_file_id=output_file_id,
            make_output_public=make_output_public,
            options=options,
        )
        return self.client.post("plugin/instance/generate", req, expect=GenerateResponse)

    def delete(self) -> PluginInstance:
        req = DeleteRequest(id=self.id)
        return self.client.post("plugin/instance/delete", payload=req, expect=PluginInstance)

    def train(
        self,
        training_request: TrainingParameterPluginInput = None,
        training_epochs: Optional[int] = None,
        export_query: Optional[str] = None,
        testing_holdout_percent: Optional[float] = None,
        test_split_seed: Optional[int] = None,
        training_params: Optional[Dict] = None,
        inference_params: Optional[Dict] = None,
    ) -> Task[TrainPluginOutput]:
        """Train a plugin instance.  Please provide either training_request OR the other parameters; passing
        training_request ignores all other parameters, but is kept for backwards compatibility.
        """
        input_params = training_request or TrainingParameterPluginInput(
            plugin_instance=self.handle,
            training_epochs=training_epochs,
            export_plugin_input=ExportPluginInput(
                plugin_instance=SIGNED_URL_EXPORTER_INSTANCE_HANDLE, type="file", query=export_query
            ),
            testing_holdout_percent=testing_holdout_percent,
            test_split_seed=test_split_seed,
            training_params=training_params,
            inference_params=inference_params,
        )
        return self.client.post(
            "plugin/instance/train",
            payload=input_params,
            expect=TrainPluginOutput,
        )

    def get_training_parameters(
        self, training_request: TrainingParameterPluginInput
    ) -> TrainingParameterPluginOutput:
        return self.client.post(
            "plugin/instance/getTrainingParameters",
            payload=training_request,
            expect=TrainingParameterPluginOutput,
        )

    def refresh_init_status(self):
        new_self = PluginInstance.get(self.client, handle=self.handle)
        self.init_status = new_self.init_status

    def wait_for_init(
        self,
        max_timeout_s: float = 180,
        retry_delay_s: float = 1,
    ):
        """Polls and blocks until the init has succeeded or failed (or timeout reached).

        Parameters
        ----------
        max_timeout_s : int
            Max timeout in seconds. Default: 180s. After this timeout, an exception will be thrown.
        retry_delay_s : float
            Delay between status checks. Default: 1s.
        """
        t0 = time.perf_counter()
        refresh_count = 0
        while (
            time.perf_counter() - t0 < max_timeout_s
            and self.init_status == InvocableInitStatus.INITIALIZING
        ):
            time.sleep(retry_delay_s)
            self.refresh_init_status()
            refresh_count += 1

        # If the task did not complete within the timeout, throw an error
        if self.init_status == InvocableInitStatus.INITIALIZING:
            raise SteamshipError(
                message=f"Plugin Instance {self.id} did not complete within requested timeout of {max_timeout_s}s. The init is still running on the server. You can retrieve its status via PluginInstance.get() or try waiting again with wait_for_init()."
            )



import abc
import contextlib
import logging
import uuid
from abc import ABC
from typing import Any, Dict, List, Optional, Type, cast

from steamship import Block, Steamship, Task
from steamship.agents.logging import AgentLogging
from steamship.agents.schema import AgentContext, Tool
from steamship.agents.service.agent_service import AgentService
from steamship.data.workspace import SignedUrl, Workspace
from steamship.invocable.dev_logging_handler import DevelopmentLoggingHandler
from steamship.utils.signed_urls import upload_to_signed_url


class SteamshipREPL(ABC):
    """Base class for building REPLs that facilitate running Steamship code in the IDE."""

    client: Steamship
    dev_logging_handler: DevelopmentLoggingHandler

    def __init__(self, log_level=None):
        logger = logging.getLogger()
        logger.handlers.clear()
        logger.setLevel(log_level or logging.DEBUG)
        dev_logging_handler = DevelopmentLoggingHandler()
        logger.addHandler(dev_logging_handler)

    def _make_public_url(self, block):
        filepath = str(uuid.uuid4())
        signed_url = (
            self.client.get_workspace()
            .create_signed_url(
                SignedUrl.Request(
                    bucket=SignedUrl.Bucket.PLUGIN_DATA,
                    filepath=filepath,
                    operation=SignedUrl.Operation.WRITE,
                )
            )
            .signed_url
        )
        logging.info(f"Got signed url for uploading block content: {signed_url}")
        read_signed_url = (
            self.client.get_workspace()
            .create_signed_url(
                SignedUrl.Request(
                    bucket=SignedUrl.Bucket.PLUGIN_DATA,
                    filepath=filepath,
                    operation=SignedUrl.Operation.READ,
                )
            )
            .signed_url
        )
        upload_to_signed_url(signed_url, block.raw())
        return read_signed_url

    def print_blocks(self, blocks: List[Block], metadata: Dict[str, Any]):
        """Print a list of blocks to console."""
        output = None

        for block in blocks:
            if isinstance(block, dict):
                block = Block.parse_obj(block)
            if block.is_text():
                output = block.text
            elif block.url:
                output = block.url
            elif block.content_url:
                output = block.content_url
            else:
                url = self._make_public_url(block)
                output = url

        if output:
            logging.info(
                f"{output}",
                extra={
                    AgentLogging.IS_MESSAGE: True,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )

    @contextlib.contextmanager
    def temporary_workspace(self) -> Steamship:
        workspace = Workspace.create(client=self.client)
        temp_client = Steamship(workspace=workspace.handle)
        yield temp_client
        workspace.delete()

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError()


class ToolREPL(SteamshipREPL):
    tool: Tool
    client = Steamship

    def __init__(self, tool: Tool, client: Optional[Steamship] = None):
        super().__init__()
        self.tool = tool
        self.client = client or Steamship()

    def run_with_client(self, client: Workspace, context: Optional[AgentContext] = None):
        try:
            from termcolor import colored  # noqa: F401
        except ImportError:

            def colored(message: str, color: str):
                print(message)

        if context is None:
            context = AgentContext()
        context.client = client

        print(f"Starting REPL for Tool {self.tool.name}...")
        print("If you make code changes, restart this REPL. Press CTRL+C to exit at any time.\n")

        while True:
            input_text = input(colored("Input: ", "blue"))  # noqa: F821
            input_block = Block(text=input_text)
            output = self.tool.run([input_block], context=context)
            if isinstance(output, Task):
                # TODO: Iterate on task support.
                print(f"Task: {output.task_id}")
            else:
                blocks = cast(List[Block], output)
                self.print_blocks(blocks, {})

    def run(self):
        with self.temporary_workspace() as client:
            self.run_with_client(client)


class AgentREPL(SteamshipREPL):
    agent_class: Type[AgentService]
    client = Steamship
    config = None

    def __init__(
        self,
        agent_class: Type[AgentService],
        method: str,
        agent_package_config: Optional[Dict[str, Any]],
        client: Optional[Steamship] = None,
    ):
        super().__init__()
        self.agent_class = agent_class
        self.method = method
        self.client = client or Steamship()
        self.config = agent_package_config

    def run_with_client(self, client: Steamship):
        try:
            from termcolor import colored  # noqa: F401
        except ImportError:

            def colored(text: str, color: str):
                print(text)

        print("Starting REPL for Agent...")
        print("If you make code changes, restart this REPL. Press CTRL+C to exit at any time.\n")

        agent_service = self.agent_class(client=client, config=self.config)

        while True:
            input_text = input(colored(text="Input: ", color="blue"))  # noqa: F821
            responder = getattr(agent_service, self.method)
            response = responder(input_text)
            print(colored(text=f"{response}", color="green", force_color=True))

    def run(self):
        with self.temporary_workspace() as client:
            self.run_with_client(client)



from typing import Any, List, Optional, Union

from pydantic import BaseModel

from steamship import Block, Task
from steamship.agents.schema.tool import AgentContext, Tool


class Action(BaseModel):
    """Actions represent a binding of a Tool to the inputs supplied to the tool.

    Upon completion, the Action also contains the output of the Tool given the inputs.
    """

    tool: Tool
    """Tool used by this action."""

    input: List[Block]
    """Data provided directly to the Tool (full context is passed in run)."""

    output: Optional[List[Block]] = []
    """Any direct output produced by the Tool."""


class AgentTool(Tool):
    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        pass

    name = "Agent"
    human_description = "Placeholder tool object for FinishAction"
    agent_description = ""  # intentionally left blank


class FinishAction(Action):
    """Represents a final selected action in an Agent Execution."""

    tool: Tool = AgentTool()  # Tool is hard-bound to allow storage in completed_steps in a context.
    input: List[Block] = []



from typing import Optional

from steamship.agents.tools.text_generation.text_rewrite_tool import TextRewritingTool
from steamship.utils.repl import ToolREPL

DEFAULT_LANGUAGE = """French"""

DEFAULT_PROMPT = """Instructions:
Please translate the following passage into {language}.

Passage
=======
{{input}}

{language} Translation
======================:"""


class TextTranslationTool(TextRewritingTool):
    """
    Example tool to illustrate rewriting a statement according to a particular personality.
    """

    name: str = "TextTranslationTool"
    human_description: str = "Translates a text into a new language."
    agent_description: str = "(set dynamically)"
    language: str = DEFAULT_LANGUAGE
    rewrite_prompt: str = DEFAULT_PROMPT

    def __init__(
        self, language: Optional[str] = None, rewrite_prompt: Optional[str] = None, **kwargs
    ):
        _rewrite_prompt = rewrite_prompt or DEFAULT_PROMPT
        _language = language or DEFAULT_LANGUAGE

        kwargs["rewrite_prompt"] = kwargs.get(
            "rewrite_prompt", _rewrite_prompt.format(language=_language)
        )
        kwargs["ai_description"] = kwargs.get(
            "ai_description",
            f"Used to translate text into {_language}. Input is text. Output is the text translated to {_language}.",
        )
        super().__init__(**kwargs)


if __name__ == "__main__":
    tool = TextTranslationTool(language="Spanish")
    ToolREPL(tool).run()
